import numpy as np
import enum

from pymorton import interleave2


class Form(enum.Enum):
	MATRIX_FORM = 0
	INDEXED_MATRIX_FORM = 1
	BINARY_MATRIX_FORM = 2


class Task:
	def __init__(self, code, index):
		self.code = code
		self.index = index


class Problem:
	"""
	A problem of finding an optimal alignment between two sequences \(s\) and \(t\).
	"""

	def __init__(self, shape, s=None, t=None, direction="maximize", dtype=np.float32):
		"""

		Parameters
		----------
		shape : tuple
			The shape \( (|s|, |t|) \) of the problem
		s : array_like, optional
			The actual elements of the first sequence \(s\)
		t : array_like, optional
			The actual elements of the second sequence \(t\)
		direction : {'minimize', 'maximize'}, optional
			Direction in which to optimize
		dtype : type, optional
			dtype used for computing problem affinity (or distance) values
		"""

		self._shape = tuple(shape)
		self._direction = direction
		self._dtype = dtype

		if s is not None and shape[0] != len(s):
			raise ValueError(f"sequence s [{len(s)}] does not match matrix shape {shape}")
		if t is not None and shape[1] != len(t):
			raise ValueError(f"sequence t [{len(t)}] does not match matrix shape {shape}")

		self._s = s
		self._t = t

	@property
	def shape(self):
		"""problem's shape as \(|s|, |t|\)"""

		return self._shape

	@property
	def direction(self):
		"""problem's direction, i.e. either 'maximize' or 'minimize'"""

		return self._direction

	@property
	def dtype(self):
		"""dtype used for computing problem affinity (or distance) values"""

		return self._dtype

	@property
	def s(self):
		"""elements in the first sequence \( s \) or None if not available"""

		return self._s

	@property
	def t(self):
		"""elements in the second sequence \( t \) or None if not available"""

		return self._t

	def build_matrix(self, out):
		"""
		Build a matrix M that describes an alignment problem for the two sequences
		\( s \) and \( t \). Depending on the Problem's, `direction` \( M_{i, j} \)
		contains either the affinity (or distance) between \( s_i \) and \( t_j \).

		Parameters
		----------
		out : array_like
			A suitably shaped matrix that receives \( M \).

		Returns
		-------

		"""
		raise NotImplementedError()

	@property
	def matrix(self):
		"""
		Returns
		-------
		The matrix \( M \) built by `self.build_matrix`.
		"""

		m = np.empty(self.shape, dtype=self._dtype)
		self.build_matrix(m)
		return m

	@property
	def form(self) -> Form:
		"""the sub form the problem is posed as"""

		raise NotImplementedError()

	def __str__(self):
		return str(self.matrix)


class MatrixProblem(Problem):
	"""
	A Problem that is posed as a matrix \(M\), such that \(M_{i,j}\)
	is the affinity (or distance) between the sequence elements \(s_i\)
	and \(t_j\).
	"""

	@property
	def form(self):
		return Form.MATRIX_FORM


class IndexedMatrixProblem(Problem):
	"""
	A Problem that is posed as a matrix \(M\) and two index vectors
	\(A, B\) such that \(M_{A_i,B_j}\) is the affinity (or distance)
	between the sequence elements \(s_i\) and \(t_j\).
	"""

	def build_matrix(self, out):
		sim = self.similarity_lookup_table()
		a = np.empty((self.shape[0],), dtype=np.uint32)
		b = np.empty((self.shape[1],), dtype=np.uint32)
		self.build_index_sequences(a, b)
		out[:, :] = sim[np.ix_(a, b)]

	def similarity_lookup_table(self):
		raise NotImplementedError()

	def build_index_sequences(self, a, b):
		raise NotImplementedError()

	@property
	def form(self):
		return Form.INDEXED_MATRIX_FORM


class BinaryMatrixProblem(Problem):
	def binary_similarity_values(self):
		raise NotImplementedError()

	@property
	def form(self):
		return Form.BINARY_MATRIX_FORM


class ProblemBatch:
	def __init__(self, problems, indices, master_shape):
		shapes = np.array([p.shape for p in problems])
		self._max_shape = (np.max(shapes[:, 0]), np.max(shapes[:, 1]))

		self._shape = master_shape
		self._problems = problems
		self._indices = indices

	@property
	def shape(self):
		return self._shape

	@property
	def problems(self):
		return self._problems

	@property
	def indices(self):
		return self._indices


class ProblemBag:
	def __init__(self, problems):
		self._problems = problems

		self._direction = problems[0].direction
		if not all(p.direction == self._direction for p in problems):
			raise ValueError("problems in a bag need to have same direction")

		self._dtype = problems[0].dtype
		if not all(p.dtype == self._dtype for p in problems):
			raise ValueError("problems in a bag need to have same dtype")

		self._form = Form(min(p.form.value for p in problems))

		shapes = np.array([p.shape for p in problems])
		self._max_shape = (np.max(shapes[:, 0]), np.max(shapes[:, 1]))

	def __len__(self):
		return len(self._problems)

	@property
	def problems(self):
		return self._problems

	@property
	def max_shape(self):
		return self._max_shape

	@property
	def direction(self):
		return self._direction

	@property
	def dtype(self):
		return self._dtype

	@property
	def form(self):
		return self._form

	def batches(self, batch_size):
		bag_problems = self._problems

		if batch_size == 1:
			for i, p in enumerate(bag_problems):
				yield ProblemBatch([p], [i], p.shape)
		else:
			tasks = [
				Task(interleave2(*p.shape), i) for i, p in enumerate(bag_problems)]

			sorted_i = sorted(
				np.arange(len(tasks)),
				key=lambda i: tasks[i].code)

			while True:
				while sorted_i and tasks[sorted_i[-1]].index is None:
					sorted_i.pop()

				if not sorted_i:
					break

				task = tasks[sorted_i.pop()]
				master_shape = bag_problems[task.index].shape
				indices = [task.index]

				k = 1
				while k <= len(sorted_i):
					i = sorted_i[-k]
					k += 1

					task = tasks[i]
					if task.index is None:
						continue

					s = bag_problems[task.index].shape
					if s[0] <= master_shape[0] and s[1] <= master_shape[1]:
						indices.append(task.index)
						task.index = None

						if len(indices) == batch_size:
							break

				yield ProblemBatch(
					[bag_problems[i] for i in indices],
					indices,
					master_shape)



from pyalign.tests import TestCase
from typing import Iterator, List

import pyalign.problems
import pyalign.solve
import pyalign.gaps


class TestGotoh(TestCase):
	# the following example is taken from the slides
	# about Gotoh by Rolf Backofen.

	def test_CC_ACCT(self):

		for pf in self._problems(
			"ACT",
			pyalign.problems.Equality(eq=0, ne=1),
			direction="minimize"):

			solver = pyalign.solve.GlobalSolver(
				gap_cost=pyalign.gaps.AffineGapCost(5, 1),
				codomain=Iterator[pyalign.solve.Alignment])

			problem = pf.new_problem("CC", "ACCT")
			alignments = solver.solve(problem)

			self._check_alignments(
				alignments,
				7.0,
				[[0, 0], [1, 1]],
				[[0, 2], [1, 3]]
			)

	# the following test cases are inspired from the examples given in
	# the preprint by Flouri et al., 2015:
	# "Are all global alignment algorithms and implementations correct?"

	def test_GGTGTGA_TCGCGT(self):
		'''
		from Bio import pairwise2
		distances = {
			("T", "C"): -1,
			("A", "T"): 0,
			("G", "T"): -2,
			("A", "A"): 6,
			("T", "T"): 6,
			("G", "G"): 6
		}
		def d(x, y):
			z = distances.get((x, y), distances.get((y, x), 0))
			return z
		alignments = pairwise2.align.globalcd(
			"GGTGTGA", "TCGCGT", d, -11, -1, -11, -1)
		alignments		
		'''

		for pf in self._problems(
			"ACGT",
			pyalign.problems.Dict({
				("T", "C"): -1,
				("A", "T"): 0,
				("G", "T"): -2,
				("A", "A"): 6,
				("T", "T"): 6,
				("G", "G"): 6
			}),
			direction="maximize"):

			solver = pyalign.solve.GlobalSolver(
				gap_cost=pyalign.gaps.AffineGapCost(11, 1),
				codomain=List[pyalign.solve.Alignment])

			problem = pf.new_problem("GGTGTGA", "TCGCGT")
			alignments = solver.solve(problem)

			self._check_alignments(
				alignments,
				-2,
				[[0, 0],
				 [1, 1],
				 [3, 2],
				 [4, 3],
				 [5, 4],
				 [6, 5]])

	def test_AAAGGG_TTAAAAGGGGTT(self):
		'''
		from Bio import pairwise2
		pairwise2.align.globalcd(
			"AAAGGG", "TTAAAAGGGGTT", lambda x, y: 0 if x == y else -1, -5, -1, -5, -1)
		'''

		for pf in self._problems(
			"AGT",
			pyalign.problems.Equality(eq=0, ne=-1),
			direction="maximize"):

			solver = pyalign.solve.GlobalSolver(
				gap_cost=pyalign.gaps.AffineGapCost(5, 1),
				codomain=Iterator[pyalign.solve.Alignment])

			problem = pf.new_problem("AAAGGG", "TTAAAAGGGGTT")
			alignments = solver.solve(problem)

			self._check_alignments(
				alignments,
				-14,
				[[0, 3],
				 [1, 4],
				 [2, 5],
				 [3, 6],
				 [4, 7],
				 [5, 8]],
				[[0, 0],
				 [1, 1],
				 [2, 2],
				 [3, 9],
				 [4, 10],
				 [5, 11]])

	def test_TAAATTTGC_TCGCCTTAC(self):
		'''
		from Bio import pairwise2
		pairwise2.align.globalcd(
		    "TAAATTTGC", "TCGCCTTAC", lambda x, y: 10 if x == y else -30, -40, -1, -40, -1)
		'''

		from typing import List

		for pf in self._problems(
			"ACGT",
			pyalign.problems.Equality(eq=10, ne=-30),
			direction="maximize"):

			solver = pyalign.solve.GlobalSolver(
				gap_cost=pyalign.gaps.AffineGapCost(40, 1),
				codomain=List[pyalign.solve.Alignment])

			problem = pf.new_problem("TAAATTTGC", "TCGCCTTAC")
			alignments = solver.solve(problem)

			self._check_alignments(
				alignments,
				-60,
				[[0, 0],
				 [1, 7],
				 [8, 8]],
				[[0, 6],
				 [1, 7],
				 [8, 8]])

	def test_AGAT_CTCT(self):
		'''
		from Bio import pairwise2
		pairwise2.align.globalcd(
			"AGAT", "CTCT", lambda x, y: 10 if x == y else -30, -25, -1, -25, -1)
		'''

		for open_cost, expected_score in ((25, -44), (30, -54)):

			for pf in self._problems(
				"ACGT",
				pyalign.problems.Equality(eq=10, ne=-30),
				direction="maximize"):

				solver = pyalign.solve.GlobalSolver(
					gap_cost=pyalign.gaps.AffineGapCost(open_cost, 1),
					codomain=List[pyalign.solve.Alignment])

				problem = pf.new_problem("AGAT", "CTCT")
				alignments = solver.solve(problem)

				self._check_alignments(
					alignments,
					expected_score,
					[[3, 3]])



import importlib
import importlib.util
import os
import logging
import inspect
import pyalign


def has_avx2():
	import cpufeature
	return cpufeature.CPUFeature["AVX2"]


def has_apple_m1():
	import cpuinfo
	import re
	brand = cpuinfo.get_cpu_info().get('brand_raw')
	return re.match("^Apple M1", brand) is not None


def import_algorithm():
	if os.environ.get('PYALIGN_PDOC') is not None:
		return None

	candidates = (
		('native', lambda: True),
		('intel_avx2', has_avx2),
		('apple_m1', has_apple_m1),
		('generic', lambda: True)
	)

	missing_module_names = []

	for name, check in candidates:
		module_name = f"pyalign.algorithm.{name}"
		if importlib.util.find_spec(module_name) is not None:
			if check():
				logging.info(f"running in {name} mode.")
				return importlib.import_module(module_name + ".algorithm")
			else:
				logging.info(f"{name} found, but not suitable.")
		else:
			missing_module_names.append(module_name)

	logging.info(f"pyalign installation is at {inspect.getfile(pyalign)}")

	raise RuntimeError(f"none of {missing_module_names} is available")



from pyalign.problems.instance import MatrixProblem, IndexedMatrixProblem, BinaryMatrixProblem
from pyalign.problems.function import *
from pyalign.problems.function import Function

from typing import Sequence


class Problem(MatrixProblem):
	"""
	A Problem induced by some binary function \(f(x, y)\).
	"""

	def __init__(self, f, s, t, **kwargs):
		super().__init__((len(s), len(t)), s, t, **kwargs)
		self._f = f

	def build_matrix(self, out):
		for i, x in enumerate(self.s):
			for j, y in enumerate(self.t):
				out[i, j] = self._f(x, y)


class ProblemFactory:
	"""
	A factory for problems that can be built from some binary function
	\(f(x, y)\).
	"""

	def __init__(self, w: callable, direction="maximize", dtype=np.float32):
		"""

		Parameters
		----------
		w : callable
			\(w(x, y)\) that gives a measure of affinity (or distance) between
			two arbitrary elements (e.g. characters) \(x\) and \(y\).
		direction : {'minimize', 'maximize'}
			direction of problems created by this factory, i.e. whether
			\(w\) gives affinity or distance
		dtype : type
			dtype of values returned by \(f\)
		"""

		if w is None and direction == "minimize":
			from scipy.spatial.distance import euclidean
			w = euclidean

		self._w = w
		self._direction = direction
		self._dtype = dtype

	def new_problem(self, s, t):
		"""
		Creates a new alignment problem for the sequences \(s\) and \(t\)

		Parameters
		----------
		s : array_like
			first sequence
		t : array_like
			second sequence

		Returns
		-------
		Problem modelling optimal alignment between \(s\) and \(t\)
		"""

		return Problem(
			self._w, s, t,
			direction=self._direction,
			dtype=self._dtype)


general = ProblemFactory


class AlphabetProblem(IndexedMatrixProblem):
	def __init__(self, matrix, encoder, *args, **kwargs):
		self._matrix = matrix
		self._encoder = encoder
		super().__init__(*args, **kwargs)

	def similarity_lookup_table(self):
		return self._matrix

	def build_index_sequences(self, a, b):
		encoder = self._encoder
		encoder.encode(self._s, out=a)
		encoder.encode(self._t, out=b)


class AlphabetEncoder:
	def __init__(self, alphabet):
		self._alphabet = tuple(set(alphabet))
		self._ids = dict((k, i) for i, k in enumerate(self._alphabet))

	def encode(self, s, out=None):
		ids = self._ids
		try:
			if out is None:
				return [ids[x] for x in s]
			else:
				if out.shape[0] != len(s) or len(out.shape) != 1:
					raise ValueError(f"expected shape ({len(s)},), got {out.shape}")
				for i, x in enumerate(s):
					out[i] = ids[x]
		except KeyError as e:
			raise ValueError(f"'{e.args[0]}' is not in alphabet")

	@property
	def alphabet(self):
		return self._alphabet


class AlphabetProblemFactory:
	"""
	A factory for alignment problems involving sequences \(s, t\) that can be
	written using a small fixed alphabet \(\Omega\) such that \(∀i: s_i \in
	\Omega\), \(∀j: t_j \in \Omega\).
	"""

	def __init__(self, alphabet: Sequence, w: callable, direction="maximize", dtype=np.float32):
		"""
		Parameters
		----------
		alphabet : Sequence
			fixed alphabet \Omega
		w : callable
			\(w(x, y)\) that gives a measure of affinity (or distance) between
			two arbitrary elements (e.g. characters) \(x\) and \(y\).
		direction : {'minimize', 'maximize'}
			direction of problems created by this factory, i.e. whether
			\(w\) gives affinity or distance
		dtype
			dtype of values returned by \(f\)
		"""

		self._encoder = AlphabetEncoder(alphabet)
		ordered_alphabet = self._encoder.alphabet
		n = len(ordered_alphabet)
		self._matrix = np.zeros((n, n), dtype=np.float32)
		if isinstance(w, Function):
			w.build_matrix(self._encoder, self._matrix)
		else:
			for i, x in enumerate(ordered_alphabet):
				for j, y in enumerate(ordered_alphabet):
					self._matrix[i, j] = w(x, y)
		self._direction = direction
		self._dtype = dtype

	def new_problem(self, s, t):
		return AlphabetProblem(
			self._matrix, self._encoder, (len(s), len(t)), s, t,
			direction=self._direction, dtype=self._dtype)


alphabetic = AlphabetProblemFactory


class IntegerEncoder:
	_supported_dtype =  {np.uint8, np.int8, np.uint16, np.int16, np.uint32, np.int32}

	def __init__(self):
		pass

	def encode(self, s, out=None):
		if isinstance(s, str):
			arr = np.frombuffer(s.encode("utf-32le"), dtype=np.uint32)
		elif isinstance(s, np.ndarray):
			if s.dtype not in IntegerEncoder._supported_dtype:
				raise ValueError(f"array dtype {s.dtype} is not supported")
			arr = s
		else:
			raise ValueError(f"expected str or numpy.array, got {type(s)}")

		if out is None:
			return arr
		else:
			if out.shape != arr.shape:
				raise ValueError(f"expected shape {arr.shape}, got {out.shape}")
			out[:] = arr


class BinaryProblem(BinaryMatrixProblem):
	def __init__(self, sim, encoder, *args, **kwargs):
		self._sim = sim
		self._encoder = encoder
		super().__init__(*args, **kwargs)

	def binary_similarity_values(self):
		return self._sim  # eq, ne

	def build_index_sequences(self, a, b):
		encoder = self._encoder
		encoder.encode(self._s, out=a)
		encoder.encode(self._t, out=b)


class BinaryProblemFactory:
	def __init__(self, eq=1, ne=-1, direction="maximize", dtype=np.float32):
		self._encoder = IntegerEncoder()
		self._sim = (eq, ne)
		self._direction = direction
		self._dtype = dtype

	def new_problem(self, s, t):
		return BinaryProblem(
			self._sim, self._encoder, (len(s), len(t)), s, t,
			direction=self._direction, dtype=self._dtype)


binary = BinaryProblemFactory



from pyalign.tests import TestCase

import pyalign.problems
import pyalign.solve
import pyalign.gaps


class TestNeedlemanWunsch(TestCase):
	def test_aatcg_aacg(self):
		# test case is taken from default settings at
		# http://rna.informatik.uni-freiburg.de/Teaching/index.jsp?toolName=Needleman-Wunsch

		for pf in self._problems(
			"ACGT",
			pyalign.problems.Equality(eq=1, ne=-1),
			direction="maximize"):

			problem = pf.new_problem("AATCG", "AACG")

			solver = pyalign.solve.GlobalSolver(
				gap_cost=pyalign.gaps.LinearGapCost(2),
				codomain=pyalign.solve.Alignment)

			alignment = solver.solve(problem)

			self._check_alignments(
				[alignment],
				2.0,
				[[0, 0], [1, 1], [3, 2], [4, 3]])



import numpy as np

from typing import Dict


class Function:
	def get(self, u, v):
		raise NotImplementedError()

	def __call__(self, u, v):
		return self.get(u, v)

	def build_matrix(self, encoder, matrix):
		raise NotImplementedError()

	@property
	def binary_similarity_values(self):
		return None


class Coalesced(Function):
	def __init__(self, *ops):
		self._ops = ops

	def get(self, u, v):
		for op in self._ops:
			x = op.get(u, v)
			if x is not None:
				return x
		return 0

	def build_matrix(self, encoder, matrix):
		for op in self._ops:
			op.build_matrix(encoder, matrix)


class Dict(Function):
	def __init__(self, pairs: Dict, default=0):
		self._dict = pairs
		self._default = default

	def get(self, u, v):
		a = self._dict.get((u, v))
		b = self._dict.get((v, u))

		if a is None and b is not None:
			return b
		elif b is None and a is not None:
			return a
		elif a is not None and b is not None:
			if a != b:
				raise ValueError(
					f"cost is not symmetric: d({u}, {v}) = {a} vs. d({v}, {u}) = {b}")
			return a
		else:
			return 0

	def build_matrix(self, encoder, matrix):
		matrix.fill(self._default)
		set = np.zeros(matrix.shape, dtype=bool)
		for uv, w in self._dict.items():
			i, j = encoder.encode(uv)
			if set[i, j]:
				if w != matrix[j, i]:
					raise ValueError(
						f"asymmetric w: w({i}, {j}) = {w} != w({j}, {i}) = {matrix[j, i]}")
			else:
				matrix[i, j] = w
				matrix[j, i] = w
				set[i, j] = True
				set[j, i] = True


class Equality(Function):
	def __init__(self, eq=1, ne=-1):
		self._eq = eq
		self._ne = ne

	def get(self, u, v):
		if u == v:
			return self._eq
		else:
			return self._ne

	def build_matrix(self, encoder, matrix):
		matrix.fill(self._ne)
		np.fill_diagonal(matrix, self._eq)

	@property
	def binary_similarity_values(self):
		return self._eq, self._ne



from collections.abc import Sequence
import numpy as np
import unittest


def to_tuple(x):
	if isinstance(x, Sequence):
		return tuple(map(to_tuple, x))
	else:
		return x


class TestCase(unittest.TestCase):
	def _problems(self, alphabet, sim, **kwargs):
		from ..problems import general, alphabetic, binary

		yield general(sim, **kwargs)
		yield alphabetic(alphabet, sim, **kwargs)

		eq_ne = sim.binary_similarity_values
		if eq_ne is not None:
			yield binary(*eq_ne, **kwargs)

	def _check_alignments(self, alignments, score, *edges, places=7):
		computed_edges = []

		for x in alignments:
			self.assertAlmostEqual(x.score, score, places=places)
			computed_edges.append(x.edges.tolist())

		true_edges = sorted(to_tuple(edges))
		computed_edges = sorted(to_tuple(computed_edges))

		self.assertEqual(true_edges, computed_edges)



from pyalign.tests import TestCase
from typing import Iterator

import pyalign.problems
import pyalign.solve
import pyalign.gaps


class TestBatch(TestCase):
	def test_simd(self):
		for pf in self._problems(
			"ACGT",
			pyalign.problems.Equality(eq=1, ne=-1),
			direction="maximize"):

			solver = pyalign.solve.GlobalSolver(
				gap_cost=pyalign.gaps.LinearGapCost(2),
				codomain=Iterator[pyalign.solve.Alignment],
				return_dup=False)

			alignments = solver.solve([
				pf.new_problem("AATCG", "AACG"),
				pf.new_problem("AATGC", "AACG"),
				pf.new_problem("AATCG", "AGTT")
			])

			self._check_alignments(
				alignments[0],
				2,
				[[0, 0], [1, 1], [3, 2], [4, 3]])

			self._check_alignments(
				alignments[1],
				0,
				[[0, 0], [1, 1], [2, 2], [3, 3]])

			self._check_alignments(
				alignments[2],
				-2,
				[[0, 0], [1, 1], [2, 2], [3, 3]],
				[[0, 0], [1, 1], [2, 2], [4, 3]])



from pyalign.tests import TestCase
from typing import Iterator

import pyalign.problems
import pyalign.solve
import pyalign.gaps


class TestWatermanSmithBeyer(TestCase):
	def test_cg_ccga(self):
		# test case is taken from default settings (for logarithmic gap settings) at
		# http://rna.informatik.uni-freiburg.de/Teaching/index.jsp?toolName=Waterman-Smith-Beyer

		for pf in self._problems(
			"ACG",
			pyalign.problems.Equality(eq=1, ne=-1),
			direction="maximize"):

			problem = pf.new_problem("CG", "CCGA")

			solver = pyalign.solve.GlobalSolver(
				gap_cost=pyalign.gaps.LogarithmicGapCost(3, 1),
				codomain=Iterator[pyalign.solve.Alignment])

			alignments = list(solver.solve(problem))

			self._check_alignments(
				alignments,
				-3.7,
				[[0, 0], [1, 1]],
				[[0, 0], [1, 3]],
				places=1)



import numpy as np
import time
import contextlib
import typing
import pyalign.io.alignment

from cached_property import cached_property
from functools import lru_cache
from pathlib import Path
from collections.abc import Sequence

from .gaps import GapCost, ConstantGapCost
from .problems import ProblemBag, Form
from .cpu import import_algorithm


algorithm = import_algorithm()


class Solution:
	def __init__(self, problem, solver, solution):
		self._problem = problem
		self._solver = solver
		self._solution = solution

	@property
	def problem(self):
		return self._problem

	@property
	def score(self):
		return self._solution.score

	@cached_property
	def alignment(self):
		if self._solution.alignment is None:
			return None
		return Alignment(
			self._problem,
			self._solver,
			self._solution.alignment)

	@property
	def shape(self):
		return self._solution.values.shape

	@cached_property
	def values(self):
		return self._solution.values

	@lru_cache(maxsize=2)
	def traceback(self, form="matrix"):
		if form == "matrix":
			return self._solution.traceback_as_matrix
		elif form == "edges":
			return self._solution.traceback_as_edges
		else:
			return ValueError(form)

	@cached_property
	def path(self):
		return self._solution.path

	@property
	def complexity(self):
		return self._solution.complexity

	def display(self, layer=0):
		import bokeh.io
		from pyalign.io.plot import TracebackPlotFactory
		f = TracebackPlotFactory(
			self._solution, self._problem, layer=layer)
		bokeh.io.show(f.create())

	def _ipython_display_(self):
		self.display()

	def export_image(self, path):
		import bokeh.io
		from pyalign.io.plot import TracebackPlotFactory
		f = TracebackPlotFactory(self._solution, self._problem)
		path = Path(path)
		if path.suffix == ".svg":
			bokeh.io.export_svg(f.create(), filename=path)
		else:
			bokeh.io.export_png(f.create(), filename=path)


class Alignment:
	"""
	A specific alignment that is part of a solution.
	"""

	def __init__(self, problem, solver, alignment):
		self._problem = problem
		self._solver = solver
		self._alignment = alignment

	@property
	def problem(self):
		return self._problem

	@property
	def solver(self):
		return self._solver

	@property
	def score(self):
		return self._alignment.score

	@property
	def s_to_t(self):
		return self._alignment.s_to_t

	@property
	def t_to_s(self):
		return self._alignment.t_to_s

	@cached_property
	def edges(self):
		s_to_t = self.s_to_t
		i = np.nonzero(s_to_t >= 0)[0]
		return np.column_stack([i, s_to_t[i]])

	def print(self):
		formatter = pyalign.io.alignment.Formatter(self)
		text = formatter.text
		if text is not None:
			print(text)

	def _repr_html_(self):
		formatter = pyalign.io.alignment.Formatter(self)
		return formatter.html


class Score:
	pass


def next_power_of_2(x):
	return 1 if x == 0 else 2 ** (int(x) - 1).bit_length()


class SolverCache:
	def __init__(self, options):
		self._options = options
		self._max_lim_s = 0
		self._max_lim_t = 0
		self._solvers = {}

	def ensure(self, len_s, len_t):
		lim_s = max(self._max_lim_s, next_power_of_2(len_s))
		lim_t = max(self._max_lim_t, next_power_of_2(len_t))

		if lim_s > self._max_lim_s or lim_t > self._max_lim_t:
			self._max_lim_s = lim_s
			self._max_lim_t = lim_t
			self._solvers = {}

	def get(self, len_s, len_t, direction, batch):
		self.ensure(len_s, len_t)
		key = (direction, batch)
		solver = self._solvers.get(key)

		if solver is None:
			options = self._options.copy()
			options['direction'] = algorithm.Direction.__members__[direction.upper()]
			options['batch'] = batch
			options['return_dup'] = options.get('return_dup', False)

			parsed_options = algorithm.create_options(options)

			solver = algorithm.create_solver(
				self._max_lim_s, self._max_lim_t, parsed_options)

			self._solvers[batch] = solver

		return solver


class Codomain:
	_1 = set([Score, Alignment, Solution])
	_n = dict(
		[(typing.Iterator[x], (x, "all", "iterator")) for x in [Score, Alignment, Solution]] +
		[(typing.List[x], (x, "all", "list")) for x in [Score, Alignment, Solution]]
	)

	def __init__(self, type):
		self._type = type

		if type in Codomain._1:
			self._base_type = type
			self._count = "one"
			container_type = None
			self._optimal = True
		else:
			base, self._count, container_type = Codomain._n.get(type)
			if base is None:
				raise ValueError(f"illegal codomain type '{type}'")
			self._base_type = base
			self._optimal = True

		self._key = (self.detail, self.count, container_type) + (("optimal",) if self._optimal else tuple())

	@property
	def type(self):
		return self._type

	def __str__(self):
		return str(self.type)

	@property
	def detail(self):
		return algorithm.Detail.__members__[self._base_type.__name__.upper()]

	@property
	def count(self):
		return algorithm.Count.__members__[self._count.upper()]

	@property
	def key(self):
		return self._key


class NoTimings:
	@contextlib.contextmanager
	def measure(self, name):
		yield


class Timings:
	def __init__(self, solver):
		self._solver = solver
		self._timings = dict()

	def __enter__(self):
		self._solver._timings = self
		return self

	def __exit__(self, type, value, traceback):
		self._solver._timings = NoTimings()

	@contextlib.contextmanager
	def measure(self, name):
		t0 = time.perf_counter_ns()
		yield
		t1 = time.perf_counter_ns()
		self._timings[name] = self._timings.get(name, 0) + (t1 - t0)

	def get(self):
		return self._timings

	def _ipython_display_(self):
		for k, t in self._timings.items():
			print(f"{k}: {t / 1000:.1f} µs")


def chunks(items, n):
	for i in range(0, len(items), n):
		yield items[i:i + n]


class Iterator:
	def __init__(self, problem, solver, iterator):
		self._problem = problem
		self._solver = solver
		self._iterator = iterator

	def __iter__(self):
		while True:
			x = self._iterator.next()
			if x is None:
				break
			yield self._element_class(self._problem, self._solver, x)


class AlignmentIterator(Iterator):
	_element_class = Alignment


class SolutionIterator(Iterator):
	_element_class = Solution


def make_list_factory(iterator):
	def to_list(*args, **kwargs):
		return list(iterator(*args, **kwargs))

	return to_list


def solver_variants(prefix):
	data = {
		(algorithm.Detail.SCORE, algorithm.Count.ONE, None, "optimal"): (
			"for_score", lambda _1, _2, x: x),

		(algorithm.Detail.ALIGNMENT, algorithm.Count.ONE, None, "optimal"): ("for_alignment", Alignment),
		(algorithm.Detail.ALIGNMENT, algorithm.Count.ALL, "iterator", "optimal"): (
			"for_alignment_iterator", AlignmentIterator),
		(algorithm.Detail.ALIGNMENT, algorithm.Count.ALL, "list", "optimal"): (
			"for_alignment_iterator", make_list_factory(AlignmentIterator)),

		(algorithm.Detail.SOLUTION, algorithm.Count.ONE, None, "optimal"): ("for_solution", Solution),
		(algorithm.Detail.SOLUTION, algorithm.Count.ALL, "iterator", "optimal"): (
			"for_solution_iterator", SolutionIterator),
		(algorithm.Detail.SOLUTION, algorithm.Count.ALL, "list", "optimal"): (
			"for_solution_iterator", make_list_factory(SolutionIterator))
	}

	return dict((k, (f"{prefix}_{v1}", v2)) for k, (v1, v2) in data.items())


class MatrixForm:
	_solvers = solver_variants("solve")

	def __init__(self, solver, codomain, bag):
		self._solver = solver
		batch_size = solver.batch_size
		max_shape = bag.max_shape
		self._matrix = np.zeros((max_shape[0], max_shape[1], batch_size), dtype=bag.dtype)
		self._len = np.zeros((2, batch_size), dtype=np.uint16)

		variant = MatrixForm._solvers.get(codomain.key)
		if variant is None:
			raise ValueError(f"codomain {codomain} is currently not supported")
		self._solve = getattr(solver, variant[0])
		self._construct = variant[1]

	def prepare(self, batch):
		shape = batch.shape
		matrix = self._matrix[:shape[0], :shape[1], :]
		p_len = self._len

		for k, p in enumerate(batch.problems):
			p.build_matrix(matrix[:, :, k])

			p_shape = p.shape
			p_len[0, k] = p_shape[0]
			p_len[1, k] = p_shape[1]

	def solve(self, batch):
		shape = batch.shape
		matrix = self._matrix[:shape[0], :shape[1], :]

		r = self._solve(matrix, self._len)
		c = self._construct

		return [
			c(problem, self._solver, x)
			for problem, x in zip(batch.problems, r)]


class AbstractIndexedMatrixForm:
	def prepare(self, batch):
		shape = batch.shape
		a = self._a
		b = self._b
		p_len = self._len

		for k, p in enumerate(batch.problems):
			p.build_index_sequences(
				a[k, :shape[0]],
				b[k, :shape[1]])

			p_shape = p.shape
			p_len[0, k] = p_shape[0]
			p_len[1, k] = p_shape[1]


class IndexedMatrixForm(AbstractIndexedMatrixForm):
	_solvers = solver_variants("solve_indexed")

	def __init__(self, solver, codomain, bag):
		self._solver = solver
		batch_size = solver.batch_size
		max_shape = bag.max_shape

		self._a = np.zeros((batch_size, max_shape[0]), dtype=np.uint32)
		self._b = np.zeros((batch_size, max_shape[1]), dtype=np.uint32)
		self._len = np.zeros((2, batch_size), dtype=np.uint16)

		self._sim = bag.problems[0].similarity_lookup_table()
		if not all(p.similarity_lookup_table() is self._sim for p in bag.problems):
			raise ValueError("similarity table must be identical for all problems in a batch")

		variant = IndexedMatrixForm._solvers.get(codomain.key)
		if variant is None:
			raise ValueError(f"codomain {codomain} is currently not supported")
		self._solve = getattr(solver, variant[0])
		self._construct = variant[1]

	def solve(self, batch):
		shape = batch.shape

		r = self._solve(
			self._a[:, :shape[0]],
			self._b[:, :shape[1]],
			self._sim, self._len)
		c = self._construct

		return [
			c(problem, self._solver, x)
			for problem, x in zip(batch.problems, r)]


class BinaryMatrixForm(AbstractIndexedMatrixForm):
	_solvers = solver_variants("solve_binary")

	def __init__(self, solver, codomain, bag):
		self._solver = solver
		batch_size = solver.batch_size
		max_shape = bag.max_shape

		self._a = np.zeros((batch_size, max_shape[0]), dtype=np.uint32)
		self._b = np.zeros((batch_size, max_shape[1]), dtype=np.uint32)
		self._len = np.zeros((2, batch_size), dtype=np.uint16)

		self._sim = bag.problems[0].binary_similarity_values()
		if not all(p.binary_similarity_values() is self._sim for p in bag.problems):
			raise ValueError("similarity must be identical for all problems in a batch")

		variant = BinaryMatrixForm._solvers.get(codomain.key)
		if variant is None:
			raise ValueError(f"codomain {codomain} is currently not supported")
		self._solve = getattr(solver, variant[0])
		self._construct = variant[1]

	def solve(self, batch):
		shape = batch.shape

		r = self._solve(
			self._a[:, :shape[0]],
			self._b[:, :shape[1]],
			*self._sim,  # eq, ne
			self._len)
		c = self._construct

		return [
			c(problem, self._solver, x)
			for problem, x in zip(batch.problems, r)]


class Solver:
	"""
	A solver that obtains solutions to alignment problems.
	"""

	_matrix_c = {
		Form.MATRIX_FORM: MatrixForm,
		Form.INDEXED_MATRIX_FORM: IndexedMatrixForm,
		Form.BINARY_MATRIX_FORM: BinaryMatrixForm
	}

	def __init__(
		self, gap_cost: GapCost = None, codomain=Solution, **kwargs):

		if codomain is None:
			codomain_obj = Codomain(Solution)
		else:
			codomain_obj = Codomain(codomain)
		if gap_cost is None:
			gap_cost = ConstantGapCost(0)

		self._codomain = codomain_obj

		self._options = dict(
			gap_cost=gap_cost,
			codomain=self._codomain,
			**kwargs)

		self._cache = SolverCache(self._options)
		self._timings = NoTimings()

		max_len_s = self._options.get("max_len_s")
		max_len_t = self._options.get("max_len_t")

		if max_len_s and max_len_t:
			self._cache.ensure(max_len_s, max_len_t)

	@property
	def gap_cost(self):
		return self._options["gap_cost"]

	def to_codomain(self, codomain, return_dup=False):
		kwargs = self._options.copy()
		kwargs['codomain'] = codomain
		kwargs['return_dup'] = return_dup
		return Solver(**kwargs)

	@property
	def codomain(self):
		"""the solver's codomain"""
		return self._codomain.type

	@property
	def batch_size(self):
		"""
		the solver's optimal batch size, i.e. the number of alignment pairs
		that can get processed in a single SIMD call on this machine.
		"""
		return self._cache.get(
			1, 1, direction="maximize", batch=True).batch_size

	def timings(self):
		return Timings(self)

	def _solve_bag(self, bag):
		max_shape = bag.max_shape
		solver = self._cache.get(
			max_shape[0], max_shape[1],
			direction=bag.direction,
			batch=len(bag) > 1)
		batch_size = solver.batch_size
		form = bag.form
		problems = bag.problems

		matrix_c = Solver._matrix_c.get(form)
		if matrix_c is None:
			raise ValueError(form)
		form_solver = matrix_c(
			solver, self._codomain, bag)

		result = [None] * len(problems)

		for batch in bag.batches(batch_size):
			with self._timings.measure("prepare"):
				form_solver.prepare(batch)

			with self._timings.measure("solve"):
				for i, r in zip(batch.indices, form_solver.solve(batch)):
					result[i] = r

		return result

	def _solve_problem(self, problem):
		batch = ProblemBag([problem])
		result = self._solve_bag(batch)
		return result[0]

	def solve(self, x):
		if isinstance(x, Sequence):
			return self._solve_bag(ProblemBag(list(x)))
		else:
			return self._solve_problem(x)


class LocalSolver(Solver):
	"""
	A solver that obtains optimal local alignments. For linear gap costs,
	this conforms to the results obtained with the Smith-Waterman algorithm.
	"""

	def __init__(self, gap_cost: GapCost = None, **kwargs):
		super().__init__(
			solver=algorithm.Type.ALIGNMENT,
			locality=algorithm.Locality.LOCAL,
			gap_cost=gap_cost, **kwargs)


class GlobalSolver(Solver):
	"""
	A solver that obtains optimal global alignments. For linear gap costs,
	this conforms to the results obtained with the Needleman-Wunsch algorithm.
	"""

	def __init__(self, gap_cost: GapCost = None, **kwargs):
		super().__init__(
			solver=algorithm.Type.ALIGNMENT,
			locality=algorithm.Locality.GLOBAL,
			gap_cost=gap_cost, **kwargs)


class SemiglobalSolver(Solver):
	"""
	A solver that obtains optimal semiglobal alignments.
	"""

	def __init__(self, gap_cost: GapCost = None, **kwargs):
		super().__init__(
			solver=algorithm.Type.ALIGNMENT,
			locality=algorithm.Locality.SEMIGLOBAL,
			gap_cost=gap_cost, **kwargs)


class ElasticSolver(Solver):
	"""
	A solver that uses dynamic time warping (DTW) to find the optimal solution of
	an elastic alignment problem.
	"""

	def __init__(self, **kwargs):
		super().__init__(solver=algorithm.Type.DTW, **kwargs)



"""
Fast and Versatile Alignments for Python.
"""

__pdoc__ = {
	'algorithm': False,
	'tests': False
}

from pyalign._version import __version__

from .gaps import *
from .solve import *
from .problems import *
from .simple import global_alignment, semiglobal_alignment, local_alignment



"""
A simple high level API, which should be sufficient for many simple use cases.
"""


from pyalign.gaps import LinearGapCost
from pyalign.solve import Alignment, GlobalSolver, SemiglobalSolver, LocalSolver
from pyalign.problems import binary, alphabetic
from typing import List
from functools import lru_cache


_codomain = {
    False: Alignment,
    True: List[Alignment]
}


@lru_cache(maxsize=8)
def _make_binary(eq, ne, direction):
    return binary(eq=eq, ne=ne, direction=direction)


@lru_cache(maxsize=8)
def _make_solver(solver_class, gap_cost=1, return_all=False):
    return solver_class(
        gap_cost=LinearGapCost(gap_cost),
        codomain=_codomain[return_all])


def _alignment(solver_class, a, b, score=None, gap_cost=1, return_all=False, **kwargs):
    if score is not None:
        if 'eq' in kwargs or 'ne' in kwargs:
            raise ValueError("argument 'score' cannot be used with 'ne', 'eq'")
        alphabet = list(set([*a, *b]))
        pf = alphabetic(
            alphabet, score, direction='maximize')
    else:
        pf = _make_binary(
            eq=kwargs.get('eq', 1),
            ne=kwargs.get('ne', 0),
            direction='maximize')

    solver = _make_solver(
        solver_class=solver_class,
        gap_cost=gap_cost,
        return_all=return_all)

    return solver.solve(pf.new_problem(a, b))


def global_alignment(*args, **kwargs):
    return _alignment(GlobalSolver, *args, **kwargs)


def semiglobal_alignment(*args, **kwargs):
    return _alignment(SemiglobalSolver, *args, **kwargs)


def local_alignment(*args, **kwargs):
    return _alignment(LocalSolver, *args, **kwargs)



import collections.abc
import string


def _seq_type(x):
	if isinstance(x, str):
		return 'text'
	elif isinstance(x, collections.abc.Sequence):
		if all(isinstance(y, str) for y in x):
			return 'text'
		elif all(hasattr(y, '_repr_html_') for y in x):
			return 'html'
		else:
			return None
	else:
		return None


def det_seq_type(s, t):
	seq_type = _seq_type(s)
	if seq_type != _seq_type(t):
		return None
	else:
		return seq_type


def _make_rows(s, t, ts, elastic=False):
	upper = []
	edges = []
	lower = []
	last_x = -1

	# flip upper, lower
	# flip s, t

	for i, x in enumerate(ts):
		if x < 0:
			edges.append(" ")
			lower.append(t[i])
			upper.append("-" if not elastic else s[max(last_x, 0)])
		else:
			for j in range(last_x + 1, x):
				upper.append(s[j])
				edges.append(" ")
				lower.append("-" if not elastic else t[i])

			upper.append(s[x])
			edges.append("|")
			lower.append(t[i])
			last_x = x

	for j in range(last_x + 1, len(s)):
		upper.append(s[j])
		edges.append(" ")
		lower.append("-" if not elastic else t[-1])

	return upper, edges, lower


class Formatter:
	_html_template = string.Template("""
		<table style="border-collapse: collapse; border-spacing:0;">
		<tr>$upper</tr>
		<tr style="background-color: #F0F0F0; padding:0;">$edges</tr>
		<tr>$lower</tr>
		</table>
		""")

	def __init__(self, alignment, style="t_to_s"):
		self._alignment = alignment
		self._style = style

	def _rows(self, s, t):
		elastic = self._alignment.solver.options["solver"] == "dtw"

		if self._style == "s_to_t":
			upper, edges, lower = _make_rows(t, s, self._alignment.s_to_t, elastic)
			return lower, edges, upper
		elif self._style == "t_to_s":
			return _make_rows(s, t, self._alignment.t_to_s, elastic)
		else:
			raise ValueError(self._style)

	@property
	def html(self):
		s = self._alignment.problem.s
		t = self._alignment.problem.t

		if s is None or t is None:
			return None

		seq_type = det_seq_type(s, t)

		if seq_type == "text":
			upper, edges, lower = self._rows(s, t)
		elif seq_type == "html":
			upper, edges, lower = self._rows(
				[x._repr_html_() for x in s],
				[x._repr_html_() for x in t])
		else:
			return None

		return Formatter._html_template.substitute(
			upper="".join([f"<td>{x}</td>" for x in upper]),
			edges="".join([f'<td style="text-align: center;">{x}</td>' for x in edges]),
			lower="".join([f"<td>{x}</td>" for x in lower]))

	@property
	def text(self):
		s = self._alignment.problem.s
		t = self._alignment.problem.t

		if s is None or t is None:
			return None

		if det_seq_type(s, t) != "text":
			return None

		upper, edges, lower = self._rows(s, t)

		return "\n".join([
			"".join(upper),
			"".join(edges),
			"".join(lower)
		])



import numpy as np
import yaml
import sys
import os
import platform

from pathlib import Path
from setuptools import setup, find_packages
from pybind11.setup_helpers import Pybind11Extension, build_ext

script_dir = Path(os.path.abspath(os.path.dirname(__file__)))

with open(script_dir / 'environment.yml') as f:
	required = yaml.safe_load(f.read())['dependencies'][-1]['pip']


is_arm = (platform.machine() == "arm64")  # Apple Silicon or ARM?

if is_arm:
	# workaround for cpufeature not available on Apple M1 as of 2021/09.
	required = list(filter(lambda s: not s.startswith("cpufeature"), required))

src_path = (script_dir / 'pyalign' / 'algorithm').resolve()
assert src_path.exists()

sources = [
	src_path / 'module.cpp'
]

include_dirs = [Path(np.get_include()), script_dir]

if sys.platform == 'darwin':
	cc = os.environ.get("CC")
	if cc and cc.startswith("gcc"):
		import pybind11.setup_helpers
		pybind11.setup_helpers.MACOS = False


def mk_ext(name, arch=None, cpu=None):
	extra_compile_args = []
	extra_link_args = []

	is_sanitize = os.environ.get('PYALIGN_SANITIZE_ADDRESS', False)

	if is_sanitize:
		is_debug_build = True
	else:
		is_debug_build = os.environ.get("PYALIGN_DEBUG_BUILD", False)

	if os.name == 'nt':
		pass
	else:
		if is_debug_build:
			extra_compile_args.append("-O0")
			extra_compile_args.append("-g")
		else:
			extra_compile_args.append("-O3")

		extra_compile_args.extend([
			"-ftemplate-backtrace-limit=0"])
		if arch is not None:
			extra_compile_args.append(f"-march={arch}")
		if cpu is not None:
			extra_compile_args.append(f"-mcpu={cpu}")

	if is_sanitize:
		extra_compile_args.append('-fsanitize=address')
		extra_compile_args.append('-fno-omit-frame-pointer')
		extra_compile_args.append('-fno-optimize-sibling-calls')
		extra_link_args.append('-fsanitize=address')

	return Pybind11Extension(
		f'pyalign.algorithm.{name}.algorithm',
		[str(x) for x in sorted(sources)],
		cxx_std=17,
		extra_compile_args=extra_compile_args,
		extra_link_args=extra_link_args,
		include_dirs=[str(x) for x in include_dirs],
	)

ext_modules = []

if os.environ.get("PYALIGN_PREBUILT_MARCH"):
	ext_modules.append(mk_ext('generic', None))
	if os.environ.get("PYALIGN_APPLE_M1"):
		ext_modules.append(mk_ext('apple_m1', cpu='apple-m1'))
	else:
		ext_modules.append(mk_ext('intel_avx2', arch='haswell'))
else:
	if os.environ.get("PYALIGN_APPLE_M1"):
		ext_modules.append(mk_ext('apple_m1', cpu='apple-m1'))
	else:
		ext_modules.append(mk_ext('native', arch='native'))

with open(script_dir / 'README.md') as f:
	long_description = f.read()

exec(open("pyalign/_version.py").read())

setup(
	name='pyalign',
	version=__version__,
	packages=find_packages(include=[
		'pyalign',
		'pyalign.algorithm',
		'pyalign.problems',
		'pyalign.gaps',
		'pyalign.io',
		'pyalign.tests']),
	python_requires='>=3.8',
	license='GPLv2',
	author='Bernhard Liebl',
	author_email='liebl@informatik.uni-leipzig.de',
	ext_modules=ext_modules,
	cmdclass={"build_ext": build_ext},
	install_requires=required,
	test_suite='nose.collector',
	tests_require=['nose'],
	description='Fast and Versatile Alignments for Python',
	long_description=long_description,
	long_description_content_type='text/markdown',
	#include_package_data=True,
)



import bokeh.plotting
import bokeh.models
import bokeh.io
import numpy as np


def inset_arrows(data, d0=0.1):
	dx = data['x_end'] - data['x_start']
	dy = data['y_end'] - data['y_start']

	mx = (data['x_start'] + data['x_end']) / 2
	my = (data['y_start'] + data['y_end']) / 2
	length = np.sqrt(np.power(dx, 2) + np.power(dy, 2))

	with np.errstate(divide='ignore'):
		d = d0 / length

	data_short = dict(
		x_start=mx - dx * d,
		y_start=my - dy * d,
		x_end=mx + dx * d,
		y_end=my + dy * d
	)

	cond = np.logical_and(np.logical_and(np.abs(dx) <= 1, np.abs(dy) <= 1), length > 0)
	return dict((k, np.where(cond, data_short[k], data[k])) for k in data.keys())


def flat_ix(a):
	return np.flip(np.dstack(np.meshgrid(
		np.arange(a.shape[1]),
		np.arange(a.shape[0]))).reshape(-1, 2), axis=-1)


class TracebackPlotFactory:
	def __init__(self, solution, problem, layer=0, cell_size=40):
		self._solution = solution
		self._problem = problem
		self._p = None
		self._len = None
		self._path = None
		self._layer = layer
		self._cell_size = cell_size

	def _create_plot(self):
		solution = self._solution

		values = solution.values[self._layer, 1:, 1:]

		len_s = values.shape[0]
		len_t = values.shape[1]
		cell_size = self._cell_size
		base_size = 50

		self._p = bokeh.plotting.figure(
			plot_width=base_size + (len_t + 1) * cell_size,
			plot_height=base_size + (len_s + 1) * cell_size,
			x_axis_location='above',
			title=None, toolbar_location=None)

		self._len = (len_s, len_t)

		path_mask = np.logical_and(
			solution.path[:, 0] >= 0,
			solution.path[:, 1] >= 0)
		self._path = solution.path[path_mask] + 1

	def _plot_grid(self):
		len_s, len_t = self._len

		len_s_1 = len_s + 1
		len_t_1 = len_t + 1
		grid_line_width = 2
		grid_color = 'lightgray'

		self._p.multi_line(
			xs=(1 + np.array([[-1, len_t]] * len_s_1) - 0.5).tolist(),
			ys=(np.repeat(1 + np.arange(0, len_s_1), 2).reshape(len_s_1, 2) - 0.5).tolist(),
			color=grid_color, line_width=grid_line_width)

		self._p.multi_line(
			xs=(np.repeat(1 + np.arange(0, len_t_1), 2).reshape(len_t_1, 2) - 0.5).tolist(),
			ys=(1 + np.array([[-1, len_s]] * len_t_1) - 0.5).tolist(),
			color=grid_color, line_width=grid_line_width)

	def _shade_optimal_path_cells(self):
		path = self._path

		source = bokeh.models.ColumnDataSource(dict(
			x=path[:, 1],
			y=path[:, 0],
			width=[1] * path.shape[0],
			height=[1] * path.shape[0]))

		glyph = bokeh.models.Rect(
			x="x", y="y", width="width", height="height",
			fill_color="orange",
			fill_alpha=0.25,
			line_color=None)
		self._p.add_glyph(source, glyph)

	def _plot_traceback_matrix_arrows(self):
		if not self._solution.traceback_has_max_degree_1:
			edges = self._solution.traceback_as_edges[self._layer]
			edges = edges[(edges >= 0).all(axis=-1).all(axis=-1)]
			src = edges[:, 0, :] + 1
			dst = edges[:, 1, :] + 1

		else:
			matrix = self._solution.traceback_as_matrix[self._layer, 1:, 1:]
			src = flat_ix(matrix) + 1
			dst = matrix.reshape(-1, 2) + 1

			path = self._path
			path_set = set(tuple(x) for x in path)
			mask = np.array([tuple(x) not in path_set for x in src], dtype=bool)
			src = src[mask]
			dst = dst[mask]

			mask = np.logical_and(dst[:, 0] >= 1, dst[:, 1] >= 1)
			src = src[mask]
			dst = dst[mask]

		arrow_color = 'blue'
		arrow_alpha = 0.5

		source = bokeh.models.ColumnDataSource(data=inset_arrows(dict(
			x_start=src[:, 1],
			y_start=src[:, 0],
			x_end=dst[:, 1],
			y_end=dst[:, 0])))
		self._p.add_layout(bokeh.models.Arrow(
			end=bokeh.models.OpenHead(
				line_color=arrow_color, line_alpha=arrow_alpha, line_width=1, size=5),
			source=source, x_start='x_start', y_start='y_start', x_end='x_end', y_end='y_end',
			line_color=arrow_color))

	def _plot_optimal_path_arrows(self):
		path = self._path
		arrow_color = 'orange'
		source = bokeh.models.ColumnDataSource(data=inset_arrows(dict(
			x_start=path[:-1, 1],
			y_start=path[:-1, 0],
			x_end=path[1:, 1],
			y_end=path[1:, 0])))
		self._p.add_layout(bokeh.models.Arrow(
			end=bokeh.models.OpenHead(line_color=arrow_color, line_width=1, size=5),
			source=source, x_start='x_start', y_start='y_start', x_end='x_end', y_end='y_end',
			line_color=arrow_color, line_width=2))

	def _plot_value_magnitudes(self):
		len_s, len_t = self._len
		solution = self._solution

		ix = flat_ix(solution.values[self._layer]) - 1

		source = bokeh.models.ColumnDataSource(
			data=dict(
				x=np.tile(np.arange(0, len_t + 1), len_s + 1),
				y=np.repeat(np.arange(0, len_s + 1), len_t + 1),
				value=[f'{x:.1f}' for x in solution.values[self._layer].flatten()],
				color=['gray' if x < 0 or y < 0 else 'black' for x, y in ix]))

		labels = bokeh.models.LabelSet(
			x='x', y='y', text='value', text_color='color',
			x_offset=0, y_offset=0, source=source, render_mode='canvas',
			text_font_size='9pt', text_align='center', text_baseline='middle')

		self._p.add_layout(labels)

	def _configure(self):
		p = self._p
		len_s, len_t = self._len

		if self._problem and isinstance(self._problem.s, str):
			p.yaxis.major_label_overrides = dict((i + 1, x) for i, x in enumerate(self._problem.s))
		if self._problem and isinstance(self._problem.t, str):
			p.xaxis.major_label_overrides = dict((i + 1, x) for i, x in enumerate(self._problem.t))

		p.xaxis.ticker = bokeh.models.FixedTicker(ticks=np.arange(0, len_t) + 1)
		p.yaxis.ticker = bokeh.models.FixedTicker(ticks=np.arange(0, len_s) + 1)

		p.y_range.flipped = True

		p.grid.grid_line_color = None

	def create(self):
		self._create_plot()
		self._plot_grid()
		self._shade_optimal_path_cells()
		self._plot_traceback_matrix_arrows()
		self._plot_optimal_path_arrows()
		self._plot_value_magnitudes()
		self._configure()
		return self._p



from pyalign.tests import TestCase
from typing import Iterator

import pyalign.problems
import pyalign.solve
import pyalign.gaps


class TestSmithWaterman(TestCase):
	def test_aatcg_aacg(self):
		# test case is taken from default settings at
		# http://rna.informatik.uni-freiburg.de/Teaching/index.jsp?toolName=Smith-Waterman

		for pf in self._problems(
			"ACGT",
			pyalign.problems.Equality(eq=1, ne=-1),
			direction="maximize"):

			problem = pf.new_problem("AATCG", "AACG")

			solver = pyalign.solve.LocalSolver(
				gap_cost=pyalign.gaps.LinearGapCost(2),
				codomain=Iterator[pyalign.solve.Alignment])

			alignments = list(solver.solve(problem))

			self._check_alignments(
				alignments,
				2,
				[[0, 0], [1, 1]],
				[[3, 2], [4, 3]])



import numpy as np


class GapCost:
	"""
	A gap cost \( w_k \) that depends only on the
	gap length \( k \).
	"""

	def to_tuple(self):
		raise NotImplementedError()

	def to_special_case(self):
		return {}

	def costs(self, n):
		raise NotImplementedError()

	@property
	def title(self):
		raise NotImplementedError()

	def _plot_matplotlib(self, ax, n):
		from matplotlib.ticker import MaxNLocator
		c = self.costs(n)
		ax.plot(c)
		ax.set_xlabel('gap length')
		ax.set_ylabel('cost')
		ax.set_ylim(-0.1, 1)
		ax.xaxis.set_major_locator(MaxNLocator(integer=True))
		ax.grid()

	def plot(self, n=None, backend="bokeh"):
		if n is None:
			n = 5

		if backend == "bokeh":
			import bokeh.plotting
			p = bokeh.plotting.figure(plot_width=600, plot_height=200)
			c = self.costs(n)
			p.line(np.arange(n), c, line_width=2)

			p.title = self.title
			p.xaxis.axis_label = 'gap length'
			p.yaxis.axis_label = 'cost'
			p.xaxis.ticker = np.arange(n)
			p.toolbar_location = None

			if np.amax(c) < 1:
				p.y_range.end = 1
			p.ray(
				x=[0], y=[1], length=0, angle=[0], color="black",
				line_width=1, line_dash='dotted')

			bokeh.plotting.show(p)
		elif backend == "matplotlib":
			import matplotlib.pyplot as plt
			fig, ax = plt.subplots(1, 1, figsize=(12, 3))
			self._plot_matplotlib(ax, n)
			fig.tight_layout()
			fig.show()
		else:
			raise ValueError(backend)

	def plot_to_image(self, fig, ax, n, format='png'):
		"""
		Produce a plot of this gap cost function.

		Parameters
		----------
		"""

		import io

		self._plot_matplotlib(ax, n)
		buf = io.BytesIO()
		fig.tight_layout()
		fig.savefig(buf, format=format)
		buf.seek(0)
		data = buf.getvalue()
		buf.close()
		return data

	def _ipython_display_(self):
		# see https://ipython.readthedocs.io/en/stable/config/integrating.html
		self.plot(5)


class ConstantGapCost(GapCost):
	"""
	A constant gap cost \( w_k = u \). \( w_k \) is the gap
	cost at length \( k \). \( u \) is a configurable parameter.
	"""

	def __init__(self, u):
		self._u = u
		assert u >= 0

	def to_tuple(self):
		return 'constant', self._u

	def to_special_case(self):
		if self._u == 0:
			return {
				'linear': 0
			}
		else:
			return {}

	def costs(self, n):
		w = np.full((n,), self._u, dtype=np.float32)
		w[0] = 0
		return w

	@property
	def title(self):
		return f"w(k) = {self._u:.1f}, w(0) = 0"


class LinearGapCost(GapCost):
	"""
	A linear gap cost \( w_k = u k \). \( w_k \) is the gap
	cost at length \( k \). \( u \) is a configurable parameter.

	Setting \( u = 0 \) effectively eliminates any gap costs.

	Notes
	-----
	   [1] Stojmirović, A., & Yu, Y.-K. (2009). Geometric Aspects of Biological Sequence
	       Comparison. Journal of Computational Biology, 16(4), 579–610. https://doi.org/10.1089/cmb.2008.0100
	"""

	def __init__(self, u):
		self._u = u
		assert u >= 0

	def to_tuple(self):
		return 'linear', self._u

	def to_special_case(self):
		return {
			'linear': self._u
		}

	def costs(self, n):
		return np.linspace(0., (n - 1) * self._u, n, dtype=np.float32)

	@property
	def title(self):
		return f"w(k) = {self._u:.1f} * k"


class AffineGapCost(GapCost):
	"""
	An affine gap cost \( w_k = u + v k \). \( w_k \) is the gap
	cost at length \( k \). \( u \) and \( v /) are configurable parameters.

	Notes
	-----
	   [1] Altschul, S. (1998). Generalized affine gap costs for protein sequence alignment.
           Proteins: Structure, 32.
	"""

	def __init__(self, open=None, extend=0, u=None, v=None):
		if open is not None:
			assert u is None and v is None
			self._u = open - extend
			self._v = extend
		else:
			assert extend == 0
			self._u = u
			self._v = v
		assert self._u >= 0
		assert self._v >= 0

	def to_tuple(self):
		return 'affine', self._u, self._v

	def to_special_case(self):
		return {
			'affine': (self._u, self._v)
		}

	def costs(self, n):
		w = self._u + np.linspace(0., (n - 1) * self._v, n, dtype=np.float32)
		w[0] = 0
		return w

	@property
	def title(self):
		return f"w(k) = {self._u:.1f} + {self._v:.1f} * k, w(0) = 0"


class LogarithmicGapCost(GapCost):
	""" A gap cost \( w_k =  u + v ln(k) \). \( w_k \) is the gap
	cost at length \( k \). \( u \) and \( v \) are configurable
	parameters.

	Notes
	-----
	   [1] Waterman, M. S. (1984). Efficient sequence alignment algorithms. Journal
	       of Theoretical Biology, 108(3), 333–337. https://doi.org/10.1016/S0022-5193(84)80037-5
	"""

	def __init__(self, u, v):
		self._u = u
		self._v = v
		assert u >= 0
		assert v >= 0

	def to_tuple(self):
		return 'logarithmic', self._u, self._v

	def costs(self, n):
		assert n > 0
		w = np.full((n,), self._u, dtype=np.float32)
		w[1:] += self._v * np.log(np.arange(1, n), dtype=np.float32)
		w[0] = 0
		return w

	@property
	def title(self):
		return f"w(k) = {self._u:.1f} + {self._v:.1f} * ln(k), w(0) = 0"


class ExponentialGapCost(GapCost):
	"""
	A gap cost \( w_k = 1 - u^{-k v} \). \( w_k \) is the gap
	cost at length \( k \). \( u \) and \( v \) are configurable
	parameters.
	"""

	def __init__(self, u, v):
		self._u = u  # base
		self._v = v  # exp term
		assert u >= 0
		assert v >= 0

	def to_tuple(self):
		return 'exponential', self._u, self._v

	def costs(self, n):
		c = np.empty((n,), dtype=np.float32)
		for i in range(n):
			c[i] = 1 - (self._u ** -(i * self._v))
		return c

	@property
	def title(self):
		return f"w(k) = 1 - {self._u:.1f}^(-k * {self._v:.1f})"


class UserFuncGapCost(GapCost):
	"""
	A custom gap cost defined by a Python user function.
	"""

	def __init__(self, costs_fn):
		self._costs_fn = costs_fn

	def to_tuple(self):
		return 'user', self._costs_fn

	def costs(self, n):
		c = np.empty((n,), dtype=np.float32)
		c[0] = 0
		for i in range(1, n):
			c[i] = self._costs_fn(i)
		return c

	@property
	def title(self):
		return f"w(k) = {self._costs_fn}, w(0) = 0"


def smooth_gap_cost(k):
	"""
	A gap cost modelled as the complement of an exponentially decaying
	function that starts at \( w_0 = 1 \) and decreases
	slowly such that \( w_k = 0.5 \) for a given \( k \).
	"""

	if k > 0:
		return ExponentialGapCost(2, 1 / k)
	else:
		return ConstantGapCost(0)
